"""`#include <linux/dirent.h>`

Not actually mentioned in the manpage for getdents.

"""
from __future__ import annotations
from rsyscall._raw import lib, ffi # type: ignore
import typing as t
import enum
from dataclasses import dataclass
from rsyscall.struct import Serializable

__all__ = [
    "DT",
    "Dirent",
    "DirentList",
    "GetdentsFileDescriptor",
]

class DT(enum.IntEnum):
    BLK = lib.DT_BLK # This is a block device.
    CHR = lib.DT_CHR # This is a character device.
    DIR = lib.DT_DIR # This is a directory.
    FIFO = lib.DT_FIFO # This is a named pipe (FIFO).
    LNK = lib.DT_LNK # This is a symbolic link.
    REG = lib.DT_REG # This is a regular file.
    SOCK = lib.DT_SOCK # This is a UNIX domain socket.
    UNKNOWN = lib.DT_UNKNOWN # The file type is unknown.

_d_name_offset = ffi.offsetof('struct linux_dirent64', 'd_name')

@dataclass
class Dirent:
    inode: int
    offset: int # the offset to seek to to see the next dirent
    type: DT
    name: str

    def __str__(self) -> str:
        return f"Dirent({self.type}, {self.name})"

    def to_bytes(self) -> bytes:
        def record(reclen: int) -> bytes:
            record = ffi.new('struct linux_dirent64*', {
                "d_ino": self.inode,
                "d_off": self.offset,
                "d_reclen": reclen,
                "d_type": self.type,
                "d_name": self.name.encode(),
            })
            data = bytes(ffi.buffer(record))
            # pad to real length
            return data + bytes(8 - (len(data) % 8))
        return record(len(record(0)))

class DirentList(t.List[Dirent], Serializable):
    def to_bytes(self) -> bytes:
        ret = b""
        for ent in self:
            ret += ent.to_bytes()
        return ret

    T = t.TypeVar('T', bound='DirentList')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        entries = []
        while len(data) > 0:
            # We do the work of from_bytes in this class instead of in Dirent because we need the
            # raw length field from the struct; merely doing len(name) will exclude padding.
            record = ffi.cast('struct linux_dirent64*', ffi.from_buffer(data))
            name_len = record.d_reclen - _d_name_offset
            # the name is padded with null bytes to make the dirent aligned,
            # so we have to use strlen to find the end
            name = ffi.string(record.d_name, name_len).decode()
            entries.append(Dirent(inode=record.d_ino, offset=record.d_off, type=DT(record.d_type), name=name))
            data = data[record.d_reclen:]
        return cls(entries)

#### Classes ####
from rsyscall.handle.fd import BaseFileDescriptor
from rsyscall.handle.pointer import Pointer, ReadablePointer

class GetdentsFileDescriptor(BaseFileDescriptor):
    async def getdents(self, dirp: Pointer[DirentList]) -> t.Tuple[ReadablePointer[DirentList], Pointer]:
        self._validate()
        with dirp.borrow(self.task) as dirp_n:
            ret = await _getdents64(self.task.sysif, self.near, dirp_n, dirp.size())
            return dirp.readable_split(ret)

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _getdents64(sysif: SyscallInterface, fd: near.FileDescriptor,
                      dirp: near.Address, count: int) -> int:
    return (await sysif.syscall(SYS.getdents64, fd, dirp, count))


#### Tests ####
from unittest import TestCase

class TestDirent(TestCase):
    def test_dirent(self) -> None:
        original = DirentList([Dirent(0, 1, DT.UNKNOWN, "hello"),
                               Dirent(124, 914, DT.SOCK, "hello"*512),
                               Dirent(0, 1, DT.UNKNOWN, "hello")])
        ret = DirentList.from_bytes(original.to_bytes())
        self.assertEqual(original, ret)



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from rsyscall.sys.resource import *

class TestResource(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = await local_thread.clone()

    async def asyncTearDown(self) -> None:
        await self.thr.exit(0)

    async def test_rlimit(self) -> None:
        old_rlimit = await (await self.thr.task.getrlimit(RLIMIT.FSIZE, await self.thr.malloc(Rlimit))).read()
        rlimit = Rlimit(old_rlimit.cur - 1, old_rlimit.max - 1)
        await self.thr.task.setrlimit(RLIMIT.FSIZE, await self.thr.ptr(rlimit))
        new_rlimit = await (await self.thr.task.getrlimit(RLIMIT.FSIZE, await self.thr.malloc(Rlimit))).read()
        self.assertEqual(rlimit, new_rlimit)
        self.assertNotEqual(old_rlimit, new_rlimit)



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread

from rsyscall.stdlib import mkdtemp
from rsyscall.sys.mount import MS
from rsyscall.sched import CLONE
from rsyscall.unistd import O
import unittest

class TestChroot(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.tmpdir = await mkdtemp(local_thread)
        self.thr = await local_thread.clone(CLONE.NEWUSER|CLONE.NEWNS)

    async def asyncTearDown(self) -> None:
        await self.tmpdir.cleanup()

    async def test_basic(self) -> None:
        await self.thr.mkdir(self.tmpdir/"proc")
        await self.thr.mount("/proc", self.tmpdir/"proc", "", MS.BIND, "")
        await self.thr.task.chroot(await self.thr.ptr(self.tmpdir))
        await self.thr.task.open(await self.thr.ptr("/proc/self"), O.RDONLY)



"`#include <sys/un.h>`"
from __future__ import annotations
import typing as t
from rsyscall._raw import ffi, lib # type: ignore
from rsyscall.sys.socket import AF, Sockaddr, _register_sockaddr
from rsyscall.path import Path
from dataclasses import dataclass
import os
from rsyscall.fcntl import O
if t.TYPE_CHECKING:
    from rsyscall.handle import FileDescriptor
    from rsyscall.thread import Thread

__all__ = [
    "PathTooLongError",
    "SockaddrUn",
]

class PathTooLongError(ValueError):
    pass

@dataclass
class SockaddrUn(Sockaddr):
    path: bytes

    family = AF.UNIX
    def __post_init__(self) -> None:
        if len(self.path) > 108:
            raise PathTooLongError("path", self.path, "is longer than the maximum unix address size")

    @staticmethod
    async def from_path(thr: Thread, path: t.Union[str, os.PathLike]) -> SockaddrUn:
        """Turn this path into a SockaddrUn, hacking around the 108 byte limit on socket addresses.

        If the passed path is too long to fit in an address, this function will open the parent
        directory with O_PATH and return SockaddrUn("/proc/self/fd/n/name").

        """
        try:
            return SockaddrUn(os.fsencode(path))
        except PathTooLongError:
            ppath = Path(path)
            fd = await thr.task.open(await thr.ram.ptr(ppath.parent), O.PATH)
            return SockaddrUnProcFd(fd, ppath.name)

    T = t.TypeVar('T', bound='SockaddrUn')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        header = ffi.sizeof('sa_family_t')
        if len(data) < header:
            raise Exception("data too smalllll", data)
        struct = ffi.cast('struct sockaddr_un*', ffi.from_buffer(data))
        cls.check_family(AF(struct.sun_family))
        if len(data) == header:
            # unnamed socket, name is empty
            length = 0
        elif struct.sun_path[0] == b'\0':
            # abstract socket, entire buffer is part of path
            length = len(data) - header
        else:
            # TODO handle the case where there's no null terminator
            # pathname socket, path is null-terminated
            length = lib.strlen(struct.sun_path)
        return cls(bytes(ffi.buffer(struct.sun_path, length)))

    def to_bytes(self) -> bytes:
        addr = ffi.new('struct sockaddr_un*', (AF.UNIX, self.path))
        real_length = ffi.sizeof('sa_family_t') + len(self.path) + 1
        return bytes(ffi.buffer(addr))[:real_length]

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct sockaddr_un')

    def __str__(self) -> str:
        return f"SockaddrUn({self.path!r})"

    async def close(self) -> None:
        pass
_register_sockaddr(SockaddrUn)

class SockaddrUnProcFd(SockaddrUn):
    def __init__(self, fd: FileDescriptor, name: str) -> None:
        super().__init__(os.fsencode(f"/proc/self/fd/{int(fd)}/{name}"))
        self.fd = fd
        self.name = name

    async def close(self) -> None:
        await self.fd.close()


#### Tests ####
from unittest import TestCase
class TestUn(TestCase):
    def test_sockaddrun(self) -> None:
        initial = SockaddrUn(b"asefliasjeflsaifje0.1")
        output = SockaddrUn.from_bytes(initial.to_bytes())
        self.assertEqual(initial, output)
        from rsyscall.sys.socket import SockaddrStorage
        out = SockaddrStorage.from_bytes(initial.to_bytes()).parse()
        self.assertEqual(initial, output)
        



"`#include <sys/resource.h>`"
from rsyscall._raw import ffi, lib # type: ignore
from rsyscall.struct import Struct
from dataclasses import dataclass
import enum
import typing as t

__all__ = [
    "PRIO",
    "RLIMIT",
    "Rlimit",
    "ResourceTask",
]

class PRIO(enum.IntFlag):
    NONE = 0
    PROCESS = lib.PRIO_PROCESS
    PGRP = lib.PRIO_PGRP
    USER = lib.PRIO_USER

class RLIMIT(enum.IntFlag):
    NONE = 0
    AS = lib.RLIMIT_AS
    CORE = lib.RLIMIT_CORE
    CPU = lib.RLIMIT_CPU
    DATA = lib.RLIMIT_DATA
    FSIZE = lib.RLIMIT_FSIZE
    LOCKS = lib.RLIMIT_LOCKS
    MEMLOCK = lib.RLIMIT_MEMLOCK
    MSGQUEUE = lib.RLIMIT_MSGQUEUE
    NICE = lib.RLIMIT_NICE
    NOFILE = lib.RLIMIT_NOFILE
    NPROC = lib.RLIMIT_NPROC
    RSS = lib.RLIMIT_RSS
    RTPRIO = lib.RLIMIT_RTPRIO
    RTTIME = lib.RLIMIT_RTTIME
    SIGPENDING = lib.RLIMIT_SIGPENDING
    STACK = lib.RLIMIT_STACK

@dataclass
class Rlimit(Struct):
    cur: int
    max: int

    def to_bytes(self) -> bytes:
        struct = ffi.new('struct rlimit*', (self.cur, self.max))
        return bytes(ffi.buffer(struct))

    T = t.TypeVar('T', bound='Rlimit')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct rlimit*', ffi.from_buffer(data))
        return cls(struct.rlim_cur, struct.rlim_max)

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct rlimit')

#### Classes ####
import rsyscall.far
from rsyscall.handle.pointer import Pointer, WrittenPointer

class ResourceTask(rsyscall.far.Task):
    async def setpriority(self, which: PRIO, prio: int) -> None:
        """set program scheduling priority

        man: setpriority(2)
        """
        return (await _setpriority(self.sysif, which, 0, prio))

    async def getpriority(self, which: PRIO) -> int:
        return (await _getpriority(self.sysif, which, 0))

    # {get,set}rlimit have a different struct rlimit from prlimit on 32-bit architectures,
    # so we annoyingly can't use the same Rlimit type for them.
    # So we'll just implement them in terms of prlimit.
    async def setrlimit(self, resource: RLIMIT, rlim: WrittenPointer[Rlimit]) -> None:
        await self.prlimit(resource, rlim, None)

    async def getrlimit(self, resource: RLIMIT, rlim: Pointer[Rlimit]) -> Pointer[Rlimit]:
        await self.prlimit(resource, None, rlim)
        return rlim

    async def prlimit(self, resource: RLIMIT,
                      new_limit: t.Optional[Pointer[Rlimit]]=None,
                      old_limit: t.Optional[Pointer[Rlimit]]=None) -> None:
        await _prlimit(self.sysif, None, resource,
                       new_limit.near if new_limit else None,
                       old_limit.near if old_limit else None)

#### Raw syscalls ####
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS
from rsyscall.near.types import (
    Address,
    Process,
)

async def _setpriority(
        sysif: SyscallInterface,
        which: PRIO,
        who: int,
        prio: int,
) -> None:
    """The raw setpriority syscall

    This syscall is tricky to assign a type to. It's similar to waitid, but has the extra
    complication that we can pass 0 for the "who" argument to operate on the current process/process
    group/user, as determined by PRIO.

    I considered making the type of "who", t.Union[Process, ProcessGroup, User, PRIO], and removing
    the "which" argument; passing PRIO would cause setpriority to operate on the current
    Process/ProcessGroup/User as determined by the PRIO value. That's the same as how our waitid
    wrapper works, but that usage of PRIO seemed too weird.

    """
    await sysif.syscall(SYS.setpriority, which, who, prio)

async def _getpriority(
        sysif: SyscallInterface,
        which: PRIO,
        who: int,
) -> int:
    return await sysif.syscall(SYS.getpriority, which, who)

async def _prlimit(
        sysif: SyscallInterface,
        pid: t.Optional[Process],
        resource: RLIMIT,
        new_limit: t.Optional[Address]=None,
        old_limit: t.Optional[Address]=None,
) -> None:
    await sysif.syscall(SYS.prlimit64, pid or 0, resource, new_limit or 0, old_limit or 0)



"`#include <sched.h>`"
from __future__ import annotations
from rsyscall._raw import lib, ffi # type: ignore
from rsyscall.struct import Serializable, Serializer, Struct, bits
from dataclasses import dataclass
import enum
import typing as t
import struct
import contextlib
if t.TYPE_CHECKING:
    from rsyscall.handle import Task, Pointer, WrittenPointer, FileDescriptor
    from rsyscall.loader import NativeFunction

__all__ = [
    'CpuSet',
    'Borrowable', 'Stack',
]

class CLONE(enum.IntFlag):
    "The flag argument to clone, unshare, and setns"
    NONE = 0
    ### other flags for clone
    VFORK = lib.CLONE_VFORK
    CHILD_CLEARTID = lib.CLONE_CHILD_CLEARTID
    ### sharing-control
    PARENT = lib.CLONE_PARENT
    VM = lib.CLONE_VM
    SIGHAND = lib.CLONE_SIGHAND
    IO = lib.CLONE_IO
    # valid for unshare
    FILES = lib.CLONE_FILES
    FS = lib.CLONE_FS
    NEWCGROUP = lib.CLONE_NEWCGROUP
    NEWIPC = lib.CLONE_NEWIPC
    NEWNET = lib.CLONE_NEWNET
    NEWNS = lib.CLONE_NEWNS
    NEWPID = lib.CLONE_NEWPID
    NEWUSER = lib.CLONE_NEWUSER
    NEWUTS = lib.CLONE_NEWUTS
    SYSVSEM = lib.CLONE_SYSVSEM


class Borrowable:
    """An interface for objects that can be borrowed.

    This doesn't make much sense in this file, but its primary use is for the
    generic type argument to Stack; we need to be able to borrow all the values
    inside the Stack, without knowing concretely what is inside.

    """
    def borrow_with(self, stack: contextlib.ExitStack, task: Task) -> None:
        "Borrow this value in this `task` on this `stack`, so that this value is valid while the borrow lives"
        raise NotImplementedError("borrow_with not implemented on", type(self))

T_borrowable = t.TypeVar('T_borrowable', bound=Borrowable)

_address = struct.Struct("Q")

@dataclass
class Stack(Serializable, t.Generic[T_borrowable]):
    """The stack argument to clone

    All rsyscall threads, after executing the clone syscall, immediately call
    the "ret" instruction. Thus, the first value on any stack passed to clone
    must be a function pointer, which ret will pop off and begin executing.

    After that, the stack can hold any arbitrary data which the function can
    use. That data may include rsyscall types, which will be borrowed at clone
    time to ensure they are valid at least for the start of the function's
    execution.

    """
    function: Pointer[NativeFunction]
    data: T_borrowable
    serializer: Serializer[T_borrowable]

    def borrow_with(self, stack: contextlib.ExitStack, task: Task) -> None:
        stack.enter_context(self.function.borrow(task))
        self.data.borrow_with(stack, task)

    def to_bytes(self) -> bytes:
        return _address.pack(int(self.function.near)) + self.serializer.to_bytes(self.data)

    T_stack = t.TypeVar('T_stack', bound='Stack')
    @classmethod
    def from_bytes(cls: t.Type[T_stack], data: bytes) -> T_stack:
        raise Exception("nay")

class CpuSet(Struct, t.Set[int]):
    """cpu_set_t, as used in sched_setaffinity and sched_getaffinity

    Currently we fix the size of the cpuset at 1024 bits, just like glibc. To
    support systems with more CPUs, we should loosen this fixed-size-ness.

    """
    def to_bytes(self) -> bytes:
        output = [0]*16
        for val in self:
            idx = val // 64
            bit = val % 64
            output[idx] |= 1 << bit
        return bytes(ffi.buffer(ffi.new('cpu_set_t const*', (output,))))

    T = t.TypeVar('T', bound='CpuSet')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('cpu_set_t*', ffi.from_buffer(data))
        ret: t.List[int] = []
        for i, val in enumerate(getattr(struct, '__bits')):
            inc = (64*i)
            for bit in bits(val, one_indexed=False):
                ret.append(bit)
        return cls(ret)

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('cpu_set_t')

import rsyscall.far
class SchedTask(rsyscall.far.Task):
    async def sched_setaffinity(self, mask: WrittenPointer[CpuSet]) -> None:
        with mask.borrow(self) as mask_n:
            await _sched_setaffinity(self.sysif, 0, mask.size(), mask_n)

    async def sched_getaffinity(self, mask: Pointer[CpuSet]) -> Pointer[CpuSet]:
        with mask.borrow(self) as mask_n:
            await _sched_getaffinity(self.sysif, 0, mask.size(), mask_n)
        return mask

    async def setns(self, fd: FileDescriptor, nstype: CLONE) -> None:
        fd._validate()
        await _setns(self.sysif, fd.near, nstype)

#### Raw syscalls ####
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS
import rsyscall.near as near

async def _clone(sysif: SyscallInterface, flags: int, child_stack: t.Optional[near.Address],
                ptid: t.Optional[near.Address], ctid: t.Optional[near.Address],
                newtls: t.Optional[near.Address]) -> near.Process:
    # We don't use CLONE_THREAD, so we can say without confusion, that clone returns a Process.
    if child_stack is None:
        child_stack = 0 # type: ignore
    if ptid is None:
        ptid = 0 # type: ignore
    if ctid is None:
        ctid = 0 # type: ignore
    if newtls is None:
        newtls = 0 # type: ignore
    return near.Process(await sysif.syscall(SYS.clone, flags, child_stack, ptid, ctid, newtls))

async def _unshare(sysif: SyscallInterface, flags: CLONE) -> None:
    await sysif.syscall(SYS.unshare, flags)

async def _sched_setaffinity(sysif: SyscallInterface, pid: int, cpusetsize: int, mask: near.Address) -> None:
    await sysif.syscall(SYS.sched_setaffinity, pid, cpusetsize, mask)

async def _sched_getaffinity(sysif: SyscallInterface, pid: int, cpusetsize: int, mask: near.Address) -> None:
    await sysif.syscall(SYS.sched_getaffinity, pid, cpusetsize, mask)

async def _setns(sysif: SyscallInterface, fd: near.FileDescriptor, nstype: CLONE) -> None:
    await sysif.syscall(SYS.setns, fd, nstype)


#### Tests ####
from unittest import TestCase
class TestEpoll(TestCase):
    def test_cpu_set(self) -> None:
        initial = CpuSet([42])
        output = CpuSet.from_bytes(initial.to_bytes())
        self.assertEqual(initial, output)



# Generated by the "nixdeps" line in setup.py
import nixdeps
import pathlib
closure = nixdeps.PackageClosure(pathlib.Path("/nix/store/j1hif8jksmdr5200i8c7v4xm24zm262d-openssh-8.4p1"), [
    "/nix/store/fnkixi37wfz5nlyzwab2l51p29793a1m-libunistring-0.9.10",
    "/nix/store/ai6j2i00rik53akq8r5pi2nqjvrqi7ky-libidn2-2.3.0",
    "/nix/store/90illc73xfs933d06daq6d41njs8yh66-glibc-2.32-37",
    "/nix/store/08qdc7hl7gxs1wbk9fynmvrr4zmpaqfl-bzip2-1.0.6.0.2",
    "/nix/store/kk5rvl6m9hc3sgmzf7fw5735k6izkbvb-attr-2.4.48",
    "/nix/store/2dx5xnd837631gcdj0z00h7xlv9jp402-acl-2.2.53",
    "/nix/store/w65ydq10abi813fi7d5j7afdcrxj3aqq-bash-4.4-p23",
    "/nix/store/5iqkifw105y41b226dqqb4279isd7bwv-audit-2.8.5",
    "/nix/store/6hx1crbjzqbjmy907p7dzdqq4sbxszp6-libcap-2.47-lib",
    "/nix/store/9fkdn82nzxgkxyax3m9qdh2knj7apaa5-coreutils-8.32",
    "/nix/store/6hhqb94ilrdmh0cx5bhdqb92m6bqkvdj-zlib-1.2.11",
    "/nix/store/b6p5cspsidkc2bpljhwz7d0nbmgrx5z0-glibc-2.32-37-bin",
    "/nix/store/fmxgxr7kx29aqbzjp662v5mhkrbjvl91-gcc-10.2.0-lib",
    "/nix/store/j4ai7j4mazjy6lnq1dbmwv98gvjmiq6f-db-4.8.30",
    "/nix/store/im1gjz24adc97l8linz1ryrzg0x1gg1l-linux-pam-1.5.1",
    "/nix/store/9afqpaqknbcqcpiq0vplm6pq9lazk5cc-shadow-4.8.1",
    "/nix/store/bb4l2lmd6vrcyy40ciac8g6wb49a8szp-util-linux-2.36.1",
    "/nix/store/a4qjxgdrsyk8lvb0pwgmjfscbpvifakw-util-linux-2.36.1-bin",
    "/nix/store/2cw85jfh8yj20cj62lcs21y5jfcq28bi-xz-5.2.5",
    "/nix/store/k5s62x4i13ri5v6q75hphahqf1jgl6g0-kmod-27",
    "/nix/store/m9fxvc4gsmprv1cir77m2i942x5jhlb2-kexec-tools-2.0.20",
    "/nix/store/4jllf2xc8cpk25cqmnfkhf3xhfdmgrjk-ncurses-6.2",
    "/nix/store/ycrv6mkzhxl0vj5xvy8hl4ad095hv42s-readline-7.0p5",
    "/nix/store/mshinwyjb1imfpdjmrpbbh3r7zz5sn6x-bash-interactive-4.4-p23",
    "/nix/store/pqlxb52qjmg4sq2m80zh27ppqr4l0zhx-libseccomp-2.5.1-lib",
    "/nix/store/89kh3d91b9i9mqh39fj6xdff1sw81m9f-gzip-1.10",
    "/nix/store/vw1lcig1bkncmsmhif27h8hq1gaxldhg-bzip2-1.0.6.0.2-bin",
    "/nix/store/rrrdjd8qmsravhwdansg969xmss8yc26-kbd-2.0.4",
    "/nix/store/gjxyg641kinpssxyd1nvbrbgd3yi5hg8-systemd-minimal-247.3",
    "/nix/store/1z35cc7x8dg2a1ygjaycqzfaqmd2bf2n-libusb-1.0.24",
    "/nix/store/0m1dvm2dayzz24k7bdjrlq2kgc2ylxrz-hidapi-0.10.1",
    "/nix/store/1pzq6yh1r415ykizk6zw61z5jlkx9994-openssl-1.1.1j",
    "/nix/store/dx296x7d0v2cdbz3ya2mrd67pq8fvd5d-libcbor-0.8.0",
    "/nix/store/1qaww47sckfihaz8z2zds5vcq3zzkjlf-libfido2-1.5.0",
    "/nix/store/z88sdrj94jhbp7v8vxf1z8xy6vcv8wln-keyutils-1.6.3-lib",
    "/nix/store/f059zv2fl9ml1q53qj5wbwi8xiim28g1-libkrb5-1.18",
    "/nix/store/m2hzy9d98gy448f8s47iqq11y37qp3x5-libedit-20210216-3.1",
    "/nix/store/j1hif8jksmdr5200i8c7v4xm24zm262d-openssh-8.4p1",
])



"`#include <time.h>`"
from __future__ import annotations
from rsyscall._raw import lib, ffi # type: ignore
from rsyscall.struct import Struct
import typing as t
from decimal import Decimal
import decimal
import math
from dataclasses import dataclass

NSEC_PER_SEC = 1_000_000_000

@dataclass
class Timespec:
    """struct timespec, as used by several time-related system-calls.

    This struct specifies time with nanosecond precision, but there's no good standard way
    in Python to represent such times. The growing standard is to use an integer number of
    nanoseconds, but that's easy to confuse with an integer number of seconds, and most
    functions don't take number-of-nanoseconds, they take number-of-seconds.

    So, this class supports conversion to and from a bunch of other formats.

    See the proposal of using Decimal to represent nanosecond timestamps:

    https://www.python.org/dev/peps/pep-0410/
    https://www.python.org/dev/peps/pep-0564/
    https://vstinner.github.io/python37-pep-564-nanoseconds.html

    The rejection of that proposal by Guido:

    https://mail.python.org/pipermail/python-dev/2012-February/116837.html
    https://bugs.python.org/issue23084

    """
    sec: int
    nsec: int

    def to_decimal(self) -> Decimal:
        raise NotImplementedError

    def to_nanos(self) -> int:
        raise NotImplementedError

    def _to_cffi_dict(self) -> t.Dict[str, int]:
        return {
            "tv_sec": self.sec,
            "tv_nsec": self.nsec,
        }

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct timespec const*', self._to_cffi_dict())))

    T = t.TypeVar('T', bound='Timespec')
    @classmethod
    def from_float(cls: t.Type[T], value: t.Union[float, Decimal, Timespec]) -> T:
        if isinstance(value, Timespec):
            return cls(value.sec, value.nsec)
        elif isinstance(value, Decimal):
            frac, i = decimal.getcontext().divmod(value, Decimal(1))
            return cls(int(i), int(frac*NSEC_PER_SEC))
        else:
            fractional, integer = math.modf(value)
            return cls(int(integer), int(fractional*NSEC_PER_SEC))

    @classmethod
    def from_nanos(cls: t.Type[T], nsec: int) -> T:
        raise NotImplementedError

    @classmethod
    def from_cffi(cls: t.Type[T], cffi_value: t.Any) -> T:
        return cls(cffi_value.tv_sec, cffi_value.tv_nsec)

    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct timespec*', ffi.from_buffer(data))
        return cls.from_cffi(struct)

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct timespec')

@dataclass
class Itimerspec(Struct):
    interval: Timespec
    value: Timespec

    def __init__(self, interval: t.Union[float, Decimal, Timespec],
                 value: t.Union[float, Decimal, Timespec]) -> None:
        self.interval = Timespec.from_float(interval)
        self.value = Timespec.from_float(interval)

    def _to_cffi_dict(self) -> t.Dict[str, t.Dict[str, int]]:
        return {
            "it_interval": self.interval._to_cffi_dict(),
            "it_value": self.value._to_cffi_dict(),
        }

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct itimerspec const*', self._to_cffi_dict())))

    T = t.TypeVar('T', bound='Itimerspec')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct itimerspec*', ffi.from_buffer(data))
        return cls(Timespec.from_cffi(struct.it_interval), Timespec.from_cffi(struct.it_value))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct itimerspec')

#### Tests ####
from unittest import TestCase
class TestTime(TestCase):
    def test_timespec(self) -> None:
        initial = Timespec.from_float(Decimal('4.1'))
        output = Timespec.from_bytes(initial.to_bytes())
        self.assertEqual(initial, output)

    def test_itimerspec(self) -> None:
        initial = Itimerspec(Timespec.from_float(Decimal('1.2')), Timespec.from_float(Decimal('3.4')))
        output = Itimerspec.from_bytes(initial.to_bytes())
        self.assertEqual(initial, output)



"An async-supporting equivalent of eval(..., 'single')"
from dataclasses import dataclass
from arepl.acompile import compile_to_awaitable, _InternalResult
import ast
import typing as t

class Result:
    pass

@dataclass
class ReturnResult(Result):
    "The statement returned a value."
    value: t.Any

@dataclass
class ExceptionResult(Result):
    "The statement raised an exception."
    exception: BaseException

@dataclass
class ExpressionResult(Result):
    "The statement was actually an expression, and evaluated to a value."
    value: t.Any

@dataclass
class FallthroughResult(Result):
    "The statement was an assignment, or pass, or something, and we've fallen through, with nothing to print."
    pass

async def eval_single(astob: ast.Interactive, global_vars: t.Dict[str, t.Any]) -> Result:
    "Compile and evaluate this snippet of AST, with these globals, and return its result"
    awaitable = compile_to_awaitable(astob, global_vars)
    try:
        val = await awaitable
    except _InternalResult as e:
        if e.is_expression:
            return ExpressionResult(e.value)
        else:
            return FallthroughResult()
    except BaseException as e:
        # We want to skip the innermost frame of the traceback, which shows "await awaitable".
        e.__traceback__ = e.__traceback__.tb_next # type: ignore
        return ExceptionResult(e)
    else:
        return ReturnResult(val)



from setuptools import setup, find_packages

setup(name='rsyscall',
      version='0.0.2',
      description='A library for making system calls remotely, through another process, which may be located on a remote host',
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: POSIX :: Linux",
      ],
      keywords='linux syscall distributed',
      url='https://github.com/catern/rsyscall',
      author='catern',
      author_email='sbaugh@catern.com',
      license='MIT',
      cffi_modules=["ffibuilder.py:ffibuilder"],
      packages=find_packages(),
      nixdeps={'rsyscall._nixdeps': ['nix', 'librsyscall', 'openssh', 'coreutils']},
      include_package_data=True,
)



"""The lowest-level interface for making syscalls

The SyscallInterface is the segment register override prefix, which is used with the
instruction to say which segment register to use for the syscall.

In terms of this metaphor: We don't know from a segment register override prefix alone
that the near pointers we are passing to an instruction are valid pointers in the segment
currently contained in the segment register.

Translating that into concrete terms: We don't know from a SyscallInterface alone that the
identifiers we are passing to a syscall match the namespaces active in the task behind the
SyscallInterface.

(The task is like the segment register, in this analogy.)

"""
from __future__ import annotations
from dataclasses import dataclass
from rsyscall.sys.syscall import SYS
import abc
import typing as t
import os
if t.TYPE_CHECKING:
    import rsyscall.handle as handle

__all__ = [
    "SyscallInterface",
    "SyscallHangup",
]

class SyscallInterface:
    """The lowest-level interface for an object which lets us send syscalls to some process.

    We send syscalls to a process, but nothing in this interface tells us anything about
    the process to which we're sending syscalls; that information is maintained in the
    Task, which contains an object matching this interface.

    This is like the segment register override prefix, with no awareness of the contents
    of the register.

    """
    @abc.abstractmethod
    async def syscall(self, number: SYS, arg1=0, arg2=0, arg3=0, arg4=0, arg5=0, arg6=0) -> int:
        """Send a syscall and wait for it to complete, throwing on error results.

        We provide a guarantee that if the syscall was sent to the process, then we will
        not return until the syscall has completed or our connection has broken.  To
        achieve this, we shield against Python coroutine cancellation while waiting for
        the syscall response.

        This guarantee is important so that our caller can deal with state changes caused
        by the syscall. If our coroutine was cancelled in the middle of a syscall, the
        result of the syscall would be discarded, and our caller wouldn't be able to
        guarantee that state changes in the process are reflected in state changes in
        Python.

        For example, a coroutine calling waitid could be cancelled; if that happened, we
        could discard a child state change indicating that the child exited. If that
        happened, future calls to waitid on that child would be invalid, or maybe return
        events for an unrelated child. We'd be completely confused.

        Instead, thanks to our guarantee, syscalls made through this method can be treated
        as atomic: They will either be submitted and completed, or not submitted at all.
        (If they're submitted and not completed due to blocking forever, that just means
        we'll never return.) There's no possibility of making a syscall, causing a
        side-effect, and never learning about the side-effect you caused.

        Since most syscalls use this method, this guarantee applies to most syscalls.

        This prevents us from cancelling an in-progress syscall if it has already been
        submitted to the process; meaning, we can't discard the result of the syscall, we
        have to wait for it.

        This may seem excessive, but the question is, what should we assume as our default?
        That syscall results can be dropped safely, or that they cannot be dropped?
        Most syscalls are side-effectful: even a simple read usually consumes data in a
        side-effectful matter, and others allocate resources which might be leaked, or can
        cause state changes. Thus, "weakening" (droppability) is not generally true for
        syscall results: syscall results cannot, in most cases, be safely ignored.

        For callers who want to preserve the ability for their coroutine to be cancelled
        even while waiting for a syscall response, the `submit_syscall` API can be used.

        Note that this Python-level cancellation protection has nothing to do with
        actually interrupting a syscall. That ability is still preserved with this
        interface; just send a signal to trigger an EINTR in the syscalling process, and
        we'll get back that EINTR as the syscall response. If you just want to be able to
        cancel deadlocked processes, you should do that. That's the true API for
        "cancellation" of syscalls on Linux.

        Likewise, if the rsyscall server dies, or we get an EOF on the syscall connection,
        or any other event causes response.receive to throw an exception, we'll still
        return that exception; so you can always fall back on killing the rsyscall server
        to stop a deadlock.

        """
        pass

    # non-syscall operations which we haven't figured out how to get rid of yet
    @abc.abstractmethod
    async def close_interface(self) -> None:
        "Close this syscall interface, shutting down the connection to the remote process."
        pass

    @abc.abstractmethod
    def get_activity_fd(self) -> t.Optional[handle.FileDescriptor]:
        """When this file descriptor is readable, it means other things want to run on this thread.

        Users of the SyscallInterface should ensure that when they block, they are
        monitoring this fd as well.

        Typically, this is the file descriptor which the rsyscall server reads for
        incoming syscalls.

        """
        pass

@dataclass
class Syscall:
    number: SYS
    arg1: t.SupportsInt
    arg2: t.SupportsInt
    arg3: t.SupportsInt
    arg4: t.SupportsInt
    arg5: t.SupportsInt
    arg6: t.SupportsInt

    def __str__(self) -> str:
        args = [self.arg1, self.arg2, self.arg3, self.arg4, self.arg5, self.arg6]
        while args and args[-1] == 0:
            args.pop()
        return f"{self.number}({','.join(map(str, args))})"

    def __repr__(self) -> str:
        return str(self)

class SyscallError(Exception):
    """Something prevents us from returning a normal result for this syscall.

    The syscall may or may not have been actually sent to the process,
    and may or may not have actually been executed.

    This is a permanent error; all future syscalls on this interface
    will also fail with a SyscallError.

    Raised by SyscallInterface.syscall.

    """
    pass

class SyscallHangup(SyscallError):
    """This syscall was sent, but we got the equivalent of a hangup when we read the result.

    We don't know if the syscall was actually executed or not.  The
    hangup may not be actually related to the syscall we sent; we'd
    also get a hangup for syscalls if the process died.

    Note that for some syscalls (exit and exec, namely), this result
    indicates success. (Although not with absolute certainty, since
    the hangup could also be unrelated in those cases.)

    Raised by SyscallInterface.syscall.

    """
    pass

class SyscallSendError(SyscallError):
    """We encountered an error when we tried to send this syscall.

    We know for sure that this syscall was not sent and was not
    executed.

    Raised by SyscallInterface.syscall.
    """
    pass

def raise_if_error(response: int) -> None:
    "Raise an OSError if this integer is in the error range for syscall return values"
    if -4095 < response < 0:
        err = -response
        raise OSError(err, os.strerror(err))

class UnusableSyscallInterface(SyscallInterface):
    async def syscall(self, number: SYS, arg1=0, arg2=0, arg3=0, arg4=0, arg5=0, arg6=0) -> int:
        raise SyscallSendError("can't send syscalls through this sysif")

    async def close_interface(self) -> None:
        pass

    def get_activity_fd(self) -> t.Optional[handle.FileDescriptor]:
        raise Exception("can't get an activity fd from this sysif")



from __future__ import annotations

from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall.tasks.stdin_bootstrap import *

from rsyscall import local_thread

from rsyscall.tests.utils import do_async_things
from rsyscall.command import Command

class TestStdinboot(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.local = local_thread
        path = await stdin_bootstrap_path_with_nix(self.local)
        self.command = Command(path, ['rsyscall-stdin-bootstrap'], {})
        self.local_child, self.remote = await stdin_bootstrap(self.local, self.command)

    async def asyncTearDown(self) -> None:
        await self.local_child.kill()

    async def test_exit(self) -> None:
        await self.remote.exit(0)

    async def test_async(self) -> None:
        await do_async_things(self, self.remote.epoller, self.remote)

    async def test_nest(self) -> None:
        child, new_thread = await stdin_bootstrap(self.remote, self.command)
        async with child:
            await do_async_things(self, new_thread.epoller, new_thread)

    async def test_nest_multiple(self) -> None:
        for i in range(5):
            child = await self.remote.clone()
            await do_async_things(self, child.epoller, child)
            await child.exit(0)



"`#include <linux/netlink.h>`"
from rsyscall._raw import ffi, lib # type: ignore
from rsyscall.sys.socket import Sockaddr, AF, _register_sockaddr
import typing as t
import enum
from dataclasses import dataclass

__all__ = [
    "NETLINK",
    "SockaddrNl",
]

class NETLINK(enum.IntEnum):
    ROUTE = lib.NETLINK_ROUTE

@dataclass
class SockaddrNl(Sockaddr):
    # not an actual process pid, but rather "port id", which is unique per netlink socket
    pid: int
    groups: int
    family = AF.NETLINK

    def to_bytes(self) -> bytes:
        struct = ffi.new('struct sockaddr_nl*', (AF.NETLINK, 0, self.pid, self.groups))
        return bytes(ffi.buffer(struct))

    T = t.TypeVar('T', bound='SockaddrNl')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        if len(data) < cls.sizeof():
            raise Exception("data too small", data)
        struct = ffi.cast('struct sockaddr_nl*', ffi.from_buffer(data))
        cls.check_family(AF(struct.nl_family))
        return cls(struct.nl_pid, struct.nl_groups)

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct sockaddr_nl')
_register_sockaddr(SockaddrNl)



# Generated by the "nixdeps" line in setup.py
import nixdeps
import pathlib
closure = nixdeps.PackageClosure(pathlib.Path("/nix/store/dx1b537ww44fy4bma5czf05l43wwh9d4-miredo-1.2.6"), [
    "/nix/store/fnkixi37wfz5nlyzwab2l51p29793a1m-libunistring-0.9.10",
    "/nix/store/ai6j2i00rik53akq8r5pi2nqjvrqi7ky-libidn2-2.3.0",
    "/nix/store/90illc73xfs933d06daq6d41njs8yh66-glibc-2.32-37",
    "/nix/store/c9v2yc541g7kr4rhd1g8ljdhgadyy7rg-libelf-0.8.13",
    "/nix/store/dfhmwqgdiycwy7a53p77pgxlq6mh0299-libmnl-1.0.4",
    "/nix/store/kj85rj6fnxr5pzpmsn3ddl0m8yyvkjrz-libnfnetlink-1.0.1",
    "/nix/store/17sr4mnf9kii98g40ijs5vw4yqig1sym-libnetfilter_conntrack-1.0.8",
    "/nix/store/w65ydq10abi813fi7d5j7afdcrxj3aqq-bash-4.4-p23",
    "/nix/store/nkyms1w5v9pc99w604hk4d0lmp11zzhh-libpcap-1.10.0",
    "/nix/store/w123mqy21rn5r1ykmrgqdg2aax4g65vh-libnftnl-1.1.9",
    "/nix/store/gjr0vzngij7m2kyk76m6s1g6p7qqx8wa-iptables-1.8.7",
    "/nix/store/fmxgxr7kx29aqbzjp662v5mhkrbjvl91-gcc-10.2.0-lib",
    "/nix/store/z0r840f3b0p4fziwwqm1waf4aq6zj3ib-db-5.3.28",
    "/nix/store/0xk9cxmnsi3zypvnpfqbhjlif8v761nj-iproute2-5.11.0",
    "/nix/store/sxrmj115nllx5sl510lfii53b49gfr08-judy-1.0.5",
    "/nix/store/dx1b537ww44fy4bma5czf05l43wwh9d4-miredo-1.2.6",
])



from __future__ import annotations
from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from arepl import *
from rsyscall.wish import serve_repls
from rsyscall.stdlib import mkdtemp
from rsyscall.sys.socket import AF, SOCK
from rsyscall.sys.un import SockaddrUn
import unittest
import typing as t

class TestREPL(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread
        self.tmpdir = await mkdtemp(self.thr, "test_stub")
        self.sock_path = self.tmpdir/"repl.sock"

    async def test_repl(self) -> None:
        sockfd = await self.thr.make_afd(await self.thr.socket(AF.UNIX, SOCK.STREAM|SOCK.NONBLOCK))
        addr = await self.thr.ram.ptr(await SockaddrUn.from_path(self.thr, self.sock_path))
        await sockfd.handle.bind(addr)
        await sockfd.handle.listen(10)
        clientfd = await self.thr.make_afd(await self.thr.socket(AF.UNIX, SOCK.STREAM|SOCK.NONBLOCK))
        await clientfd.connect(addr)
        await clientfd.write_all_bytes(b"foo = 11\n")
        await clientfd.write_all_bytes(b"return foo * 2\n")
        ret = await serve_repls(sockfd, {'locals': locals()}, int, "hello")
        self.assertEqual(ret, 22)



from __future__ import annotations

from rsyscall.tests.trio_test_case import TrioTestCase
import rsyscall.thread
from rsyscall.nix import enter_nix_container, deploy
import rsyscall._nixdeps.nix
import rsyscall._nixdeps.coreutils
from rsyscall.tasks.ssh import *
from rsyscall import local_thread

from rsyscall.unistd import SEEK
from rsyscall.signal import Sigset, HowSIG
from rsyscall.sys.mman import MFD
from rsyscall.sched import CLONE

from rsyscall.handle import FileDescriptor
from rsyscall.thread import Thread, Command
from rsyscall.command import Command
from rsyscall.monitor import AsyncChildProcess
from rsyscall.stdlib import mkdtemp

# import logging
# logging.basicConfig(level=logging.DEBUG)

async def start_cat(thread: Thread, cat: Command,
                    stdin: FileDescriptor, stdout: FileDescriptor) -> AsyncChildProcess:
    thread = await thread.clone()
    await thread.task.inherit_fd(stdin).dup2(thread.stdin)
    await thread.task.inherit_fd(stdout).dup2(thread.stdout)
    child = await thread.exec(cat)
    return child

class TestSSH(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.local = local_thread
        self.host = await make_local_ssh(self.local)
        self.local_child, self.remote = await self.host.ssh(self.local)

    async def asyncTearDown(self) -> None:
        await self.local_child.kill()

    async def test_read(self) -> None:
        [(local_sock, remote_sock)] = await self.remote.open_channels(1)
        data = b"hello world"
        await local_sock.write(await self.local.ram.ptr(data))
        valid, _ = await remote_sock.read(await self.remote.ram.malloc(bytes, len(data)))
        self.assertEqual(len(data), valid.size())
        self.assertEqual(data, await valid.read())

    async def test_connection_multiple_channels(self) -> None:
        """Test that using open_channels(n) for n > 1 produces working channels.

        Our parallelization for this function was previously buggy and broke the channels.
        """
        [
            (local_sock, remote_sock),
            *rest,
        ] = await self.remote.open_channels(10)
        data = b'foobar'
        _, remaining = await local_sock.write(await local_thread.ptr(data))
        self.assertEqual(remaining.size(), 0, msg="Got partial write")
        read_data, _ = await remote_sock.read(await self.remote.malloc(bytes, len(data)))
        self.assertEqual(data, await read_data.read())

    async def test_exec_true(self) -> None:
        true = (await deploy(self.local, rsyscall._nixdeps.coreutils.closure)).bin('true')
        await self.remote.run(true)

    async def test_exec_pipe(self) -> None:
        [(local_sock, remote_sock)] = await self.remote.open_channels(1)
        cat = (await deploy(self.local, rsyscall._nixdeps.coreutils.closure)).bin('cat')
        thread = await self.remote.clone()
        cat_side = thread.task.inherit_fd(remote_sock)
        await remote_sock.close()
        await cat_side.dup2(thread.stdin)
        await cat_side.dup2(thread.stdout)
        child_process = await thread.exec(cat)

        in_data = await self.local.ram.ptr(b"hello")
        written, _ = await local_sock.write(in_data)
        valid, _ = await local_sock.read(written)
        self.assertEqual(in_data.value, await valid.read())

    async def test_clone(self) -> None:
        thread1 = await self.remote.clone()
        thread2 = await thread1.clone()
        await thread2.exit(0)
        await thread1.exit(0)

    async def test_nest(self) -> None:
        local_child, remote = await self.host.ssh(self.remote)
        await local_child.kill()

    async def test_copy(self) -> None:
        cat = (await deploy(self.local, rsyscall._nixdeps.coreutils.closure)).bin('cat')

        local_file = await self.local.task.memfd_create(await self.local.ptr("source"))
        remote_file = await self.remote.task.memfd_create(await self.remote.ptr("dest"))

        data = b'hello world'
        await local_file.write(await self.local.ram.ptr(data))
        await local_file.lseek(0, SEEK.SET)

        [(local_sock, remote_sock)] = await self.remote.open_channels(1)

        local_child = await start_cat(self.local, cat, local_file, local_sock)
        await local_sock.close()

        remote_child = await start_cat(self.remote, cat, remote_sock, remote_file)
        await remote_sock.close()

        await local_child.check()
        await remote_child.check()

        await remote_file.lseek(0, SEEK.SET)
        read, _ = await remote_file.read(await self.remote.ram.malloc(bytes, len(data)))
        self.assertEqual(await read.read(), data)

    async def test_sigmask_bug(self) -> None:
        thread = await self.remote.clone()
        await rsyscall.thread.do_cloexec_except(
            thread, set([fd.near for fd in thread.task.fd_handles]))
        await self.remote.task.sigprocmask((HowSIG.SETMASK,
                                            await self.remote.ram.ptr(Sigset())),
                                           await self.remote.ram.malloc(Sigset))
        await self.remote.task.read_oldset_and_check()

    async def test_nix_deploy(self) -> None:
        # make it locally so that it can be cleaned up even when the
        # remote enters the container
        tmpdir = await mkdtemp(self.local)
        async with tmpdir:
            await enter_nix_container(self.local, rsyscall._nixdeps.nix.closure, self.remote, tmpdir)
            hello = (await deploy(self.remote, rsyscall._nixdeps.coreutils.closure)).bin('echo').args('hello world')
            await self.remote.run(hello)



"""Process-independent interface to Linux system calls

rsyscall provides an interface to an ever-growing subset of Linux system calls. This interface is:

- *process-independent*: all system calls are called as methods on process objects,
  which can refer to the “local” process or to other processes under our control.
- *type-safe*: many Linux API constraints, which are usually left to user code to enforce,
  are made explicit in the type system.
- *low-level*: any action which is possible with the underlying Linux APIs, is possible with rsyscall;
  nothing is forbidden or discouraged.

## `Thread`

The main entry point in rsyscall is the `Thread`.
One `Thread` exists for each process in which we can make system calls.

Like most other things called "threads",
rsyscall `Thread`s:

- are under the complete control of a single program
- exit when that program does
- can run arbitrary system calls in parallel across multiple CPUs

Unlike other "threads":

- user code, such as arbitrary Python or C, cannot run on rsyscall threads, only system calls.

The initial `Thread` is `rsyscall.tasks.local.local_thread`, which operates in the Python interpreter process.
New processes are usually created by `Thread.clone`, which returns a `ChildThread`.

A `Thread` has a number of other conventional, helpful resources,
which are almost always present.
For example, `Thread.stdin`, `Thread.stdout`, and `Thread.stderr`.

## System calls on `Task`, `FileDescriptor`, etc.

`Thread.task` points to the lower-level `Task` object.

All system calls exist either as methods on `Task`,
or as methods on objects (such as `FileDescriptor`) returned from `Task`,
which hold a reference to `Task`.

Read `rsyscall.doc.syscall_api` for an overview of the layout of the system call API.

`Task` makes syscalls using an internal instance of `rsyscall.near.sysif.SyscallInterface`.
The main implementations are `rsyscall.tasks.local.LocalSyscall` and `rsyscall.tasks.connection.SyscallConnection`.

The distinction between `Thread` and `Task` is that `Task` provides only the bare minimum functionality guaranteed by Linux;
for example, it is not guaranteed that stdin/stdout/stderr actually exist, and `Task` does not assume they do.

## Memory allocation and access

Because a process may be in a separate address space,
memory allocation, and access by reading and writing, are explicit.
This is primarily done through the methods `Thread.ptr` and `Thread.malloc`,
which return `Pointer`s.
`Pointer`s are garbage collected, so memory freeing is automatic.

`Thread` performs memory allocation and access using internal instances of
`rsyscall.memory.allocator.AllocatorInterface` and `rsyscall.memory.transport.MemoryTransport`.
The main allocator is `rsyscall.memory.allocator.UnlimitedAllocator`,
and the main memory transports are `rsyscall.tasks.local.LocalMemoryTransport`
and `rsyscall.memory.socket_transport.SocketMemoryTransport`.

## Non-blocking operations

We can make system calls in multiple processes in parallel.
But in any specific process,
only one system call can happen at a time.

If a system call blocks, the process running it can't run other system calls.
This is undesirable, so `Thread` comes with an epoll event loop,
`rsyscall.epoller.Epoller`,
so that we can wait for an epoll event and perform non-blocking system calls.

Use `Thread.make_afd` to register a `FileDescriptor` with epoll
and get a corresponding `AsyncFileDescriptor`.

`AsyncFileDescriptor` supports `AsyncFileDescriptor.read`, `AsyncFileDescriptor.write`,
`AsyncFileDescriptor.accept`, `AsyncFileDescriptor.connect`, and other such system calls.
The interfaces are identical to `FileDescriptor`;
the async implementations simply wait for the appropriate epoll event before calling the underlying system call.

Many system calls can't meaningfully be done without blocking the thread,
so use `AsyncFileDescriptor.handle` to access the underlying `FileDescriptor` to make those system calls.

For convenience,
`AsyncFileDescriptor` also has `AsyncFileDescriptor.read_some_bytes` and `AsyncFileDescriptor.write_all_bytes`,
which abstract over memory for simple use cases where efficient buffer management is not necessary.

Likewise, the `AsyncChildProcess` object wraps `rsyscall.handle.ChildProcess`
to perform non-blocking `AsyncChildProcess.waitpid` operations on child process.
One will usually never deal with `ChildProcess`,
because `ChildThread.exec` returns an `AsyncChildProcess`,
and that is the primary way to obtain child processes.

## Child processes

As mentioned above,
new processes are typically created by `Thread.clone` which returns a `ChildThread`.

As a convenience, the `Command` object bundles up an executable path, arguments, and environment variable updates.
The `ChildThread.exec` takes a `Command` and execs into it.

The `rsyscall.environ.Environment.which` method looks up an executable name in `PATH` and returns a `Command` if found;
you can use this with `Thread.environ`.

We'll often want to inherit file descriptors into child processes;
we can use `Task.inherit_fd` to get a handle for an inherited file descriptor,
and then `FileDescriptor.disable_cloexec` to allow it to be further inherited over `ChildThread.exec`.

`rsyscall.tasks` describes several other ways to get new `Thread`s,
including privilege-escalated `Thread`s, `Thread`s on remote hosts, and persistent `Thread`s.

"""
from rsyscall.thread import Thread, ChildThread
from rsyscall.command import Command
from rsyscall.path import Path
from rsyscall.handle import (
    FileDescriptor, Task,
    WrittenPointer, Pointer,
    ReadablePointer, LinearPointer,
)
from rsyscall.epoller import AsyncFileDescriptor
from rsyscall.monitor import AsyncChildProcess
from rsyscall.struct import Int32, Int64
from rsyscall.tasks.local import local_thread
from rsyscall.sys.mman import MemoryMapping

__all__ = [
    'Thread', 'ChildThread',
    'Command',
    'Task',
    'FileDescriptor',
    'AsyncFileDescriptor',
    'AsyncChildProcess',
    'local_thread',
    'Pointer', 'WrittenPointer', 'ReadablePointer', 'LinearPointer',
]



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from rsyscall.handle.pointer import UseAfterFreeError

class TestPointer(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread

    async def test_use_after_free_ptr(self) -> None:
        buf = await self.thr.malloc(bytes, 16)
        buf.free()
        with self.assertRaises(UseAfterFreeError):
            buf.near
        str(buf)

    async def test_use_after_free_allocation(self) -> None:
        buf = await self.thr.malloc(bytes, 16)
        buf.allocation.free()
        with self.assertRaises(UseAfterFreeError):
            buf.near
        buf = await self.thr.ptr(b'foo')
        buf.allocation.free()
        with self.assertRaises(UseAfterFreeError):
            buf.near
        str(buf)



"`#include <sys/wait.h>`"
from __future__ import annotations
import typing as t
from dataclasses import dataclass
from rsyscall._raw import ffi, lib # type: ignore
from rsyscall.signal import Siginfo, SIG
from rsyscall.command import Command
import enum

class IdType(enum.IntEnum):
    PID = lib.P_PID # Wait for the child whose process ID matches id.
    PGID = lib.P_PGID # Wait for any child whose process group ID matches id.
    ALL = lib.P_ALL # Wait for any child; id is ignored.

class CLD(enum.IntEnum):
    EXITED = lib.CLD_EXITED # child called _exit(2)
    KILLED = lib.CLD_KILLED # child killed by signal
    DUMPED = lib.CLD_DUMPED # child killed by signal, and dumped core
    STOPPED = lib.CLD_STOPPED # child stopped by signal
    TRAPPED = lib.CLD_TRAPPED # traced child has trapped
    CONTINUED = lib.CLD_CONTINUED # child continued by SIGCONT

class W(enum.IntFlag):
    # wait for different statuses
    EXITED = lib.WEXITED
    STOPPED = lib.WSTOPPED
    CONTINUED = lib.WCONTINUED
    # additional options
    NOHANG = lib.WNOHANG
    NOWAIT = lib.WNOWAIT
    # wait for different kinds of children
    # Note, these are only supported in waitid after Linux 4.7;
    # before that, they were only supported in wait4.
    CLONE = lib._WCLONE
    ALL = lib._WALL
    NOTHREAD = lib._WNOTHREAD


class CalledProcessError(Exception):
    "Thrown when a process exits uncleanly; like `subprocess.CalledProcessError`"
    state: ChildState
    "State of the child at exit"
    command: t.Optional[Command]
    "Optionally attached to CalledProcessError as useful information for debugging"

    def __init__(self, state: ChildState, command: Command=None) -> None:
        super().__init__(state, command)
        self.state = state
        self.command = command

@dataclass
class ChildState:
    code: CLD
    pid: int
    uid: int
    exit_status: t.Optional[int]
    sig: t.Optional[SIG]

    @staticmethod
    def make(code: CLD, pid: int, uid: int, status: int) -> ChildState:
        if code is CLD.EXITED:
            return ChildState(code, pid, uid, status, None)
        else:
            return ChildState(code, pid, uid, None, SIG(status))

    @staticmethod
    def make_from_siginfo(siginfo: Siginfo) -> ChildState:
        return ChildState.make(CLD(siginfo.code),
                               pid=siginfo.pid, uid=siginfo.uid,
                               status=siginfo.status)

    def state(self, options: W) -> bool:
        """Return true if this W option would have returned this state change

        Mainly useful for categorizing state changes into EXITED,
        STOPPED or CONTINUED.

        """
        return bool(options & {
            CLD.EXITED: W.EXITED,
            CLD.KILLED: W.EXITED,
            CLD.DUMPED: W.EXITED,
            CLD.STOPPED: W.STOPPED,
            CLD.TRAPPED: W.STOPPED,
            CLD.CONTINUED: W.CONTINUED,
        }[self.code])

    def died(self) -> bool:
        return self.code in {CLD.EXITED, CLD.KILLED, CLD.DUMPED}

    def clean(self) -> bool:
        return self.code == CLD.EXITED and self.exit_status == 0

    def check(self) -> None:
        if self.clean():
            return None
        else:
            raise CalledProcessError(self)

    def killed_with(self) -> SIG:
        """What signal was the child killed with?

        Throws if the child was not killed with a signal.

        """
        if not self.died():
            raise Exception("Child isn't dead")
        if self.sig is None:
            raise Exception("Child wasn't killed with a signal")
        return self.sig


#### Raw syscalls ####
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS
from rsyscall.near.types import (
    Address,
    Process,
    ProcessGroup,
)

async def _waitid(sysif: SyscallInterface,
                  id: t.Union[Process, ProcessGroup, None], infop: t.Optional[Address], options: int,
                  rusage: t.Optional[Address]) -> int:
    if isinstance(id, Process):
        idtype = IdType.PID
    elif isinstance(id, ProcessGroup):
        idtype = IdType.PGID
    elif id is None:
        idtype = IdType.ALL
        id = 0 # type: ignore
    else:
        raise ValueError("unknown id type", id)
    if infop is None:
        infop = 0 # type: ignore
    if rusage is None:
        rusage = 0 # type: ignore
    return (await sysif.syscall(SYS.waitid, idtype, id, infop, options, rusage))


#### Tests ####
from unittest import TestCase

class TestWait(TestCase):
    def test_child_state(self) -> None:
        state = ChildState.make_from_siginfo(Siginfo(code=CLD.EXITED, pid=1, uid=13, status=1))
        self.assertFalse(state.clean())



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread

from rsyscall.sys.socket import *
from rsyscall.sys.un import *
from rsyscall.sys.uio import IovecList
from rsyscall.fcntl import O
from rsyscall.linux.dirent import DirentList
from rsyscall.stdlib import mkdtemp

import logging
logger = logging.getLogger(__name__)

class TestSocket(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread
        self.tmpdir = await mkdtemp(self.thr)

    async def asyncTearDown(self) -> None:
        await self.tmpdir.cleanup()

    async def test_listen(self) -> None:
        sockfd = await self.thr.task.socket(AF.UNIX, SOCK.STREAM)
        addr = await self.thr.ram.ptr(await SockaddrUn.from_path(self.thr, self.tmpdir/"sock"))
        await sockfd.bind(addr)
        await sockfd.listen(10)

        clientfd = await self.thr.task.socket(AF.UNIX, SOCK.STREAM)
        await clientfd.connect(addr)
        connfd = await sockfd.accept()

    async def test_listen_async(self) -> None:
        sockfd = await self.thr.make_afd(await self.thr.socket(AF.UNIX, SOCK.STREAM|SOCK.NONBLOCK))
        addr = await self.thr.ram.ptr(await SockaddrUn.from_path(self.thr, self.tmpdir/"sock"))
        await sockfd.handle.bind(addr)
        await sockfd.handle.listen(10)

        clientfd = await self.thr.make_afd(await self.thr.socket(AF.UNIX, SOCK.STREAM|SOCK.NONBLOCK))
        await clientfd.connect(addr)
        connfd, client_addr = await sockfd.accept_addr()
        logger.info("%s, %s", addr, client_addr)
        await connfd.close()
        await sockfd.close()
        await clientfd.close()

    async def test_listen_async_accept(self) -> None:
        sockfd = await self.thr.make_afd(await self.thr.socket(AF.UNIX, SOCK.STREAM|SOCK.NONBLOCK))
        addr = await self.thr.ram.ptr(await SockaddrUn.from_path(self.thr, self.tmpdir/"sock"))
        await sockfd.handle.bind(addr)
        await sockfd.handle.listen(10)

        clientfd = await self.thr.make_afd(await self.thr.socket(AF.UNIX, SOCK.STREAM|SOCK.NONBLOCK))
        await clientfd.connect(addr)

        connfd_h, client_addr = await sockfd.accept_addr(SOCK.NONBLOCK)
        connfd = await self.thr.make_afd(connfd_h)
        logger.info("%s, %s", addr, client_addr)
        data = b"hello"
        await connfd.write_all_bytes(data)
        self.assertEqual(data, await clientfd.read_some_bytes())
        await connfd.close()
        await sockfd.close()
        await clientfd.close()

    async def test_pass_fd(self) -> None:
        fds = await (await self.thr.task.socketpair(
            AF.UNIX, SOCK.STREAM, 0,
            await self.thr.ram.malloc(Socketpair))).read()
        in_data = b"hello"

        iovec = await self.thr.ram.ptr(IovecList([await self.thr.ram.ptr(in_data)]))
        cmsgs = await self.thr.ram.ptr(CmsgList([CmsgSCMRights([fds.second])]))
        [written], [] = await fds.second.sendmsg(
            await self.thr.ram.ptr(SendMsghdr(None, iovec, cmsgs)), SendmsgFlags.NONE)

        [valid], [], hdr = await fds.first.recvmsg(
            await self.thr.ram.ptr(RecvMsghdr(None, iovec, cmsgs)), RecvmsgFlags.NONE)

        self.assertEqual(in_data, await valid.read())

        hdrval = await hdr.read()
        [[passed_fd]] = await hdrval.control.read() # type: ignore
        self.assertEqual(hdrval.name, None)
        self.assertEqual(hdrval.flags, MsghdrFlags.CMSG_CLOEXEC)


    async def test_shutdown_read(self) -> None:
        "When we shutdown(SHUT.RD) a socket and read from it, we get pending data then EOF"
        fds = await (await self.thr.task.socketpair(
            AF.UNIX, SOCK.STREAM, 0,
            await self.thr.ram.malloc(Socketpair))).read()

        data = b'hello'
        await fds.second.write(await self.thr.ptr(data))
        await fds.first.shutdown(SHUT.RD)
        read, _ = await fds.first.read(await self.thr.malloc(bytes, 4096))
        self.assertEqual(read.size(), len(data))

    async def test_long_sockaddr(self) -> None:
        "SockaddrUn.from_path works correctly on long Unix socket paths"
        longdir = await self.thr.ram.ptr(self.tmpdir/("long"*50))
        await self.thr.task.mkdir(longdir)
        addr = await self.thr.ram.ptr(await SockaddrUn.from_path(self.thr, longdir.value/"sock"))

        sockfd = await self.thr.task.socket(AF.UNIX, SOCK.STREAM)
        await sockfd.bind(addr)
        await sockfd.listen(10)
        
        clientfd = await self.thr.task.socket(AF.UNIX, SOCK.STREAM)
        await clientfd.connect(addr)
        connfd = await sockfd.accept()

        dirfd = await self.thr.task.open(longdir, O.DIRECTORY)
        valid, rest = await dirfd.getdents(await self.thr.ram.malloc(DirentList, 4096))
        self.assertCountEqual([dirent.name for dirent in await valid.read()], ['.', '..', 'sock'])



"""Pure functions to support creating REPLs which can run asynchronous code.

There's no actual REPL in here; this is all
[sans-io](https://sans-io.readthedocs.io/).

"""
from arepl.repl import PureREPL, ExpressionResult, run_repl, FromREPL



"""Filesystem-watching implemented using inotify

Nothing special here, this is just normal inotify usage.

"""
from __future__ import annotations
from dataclasses import dataclass, field
from dneio import RequestQueue, reset, Continuation
from rsyscall import Pointer
from rsyscall.epoller import AsyncFileDescriptor, AsyncReadBuffer
from rsyscall.memory.ram import RAM
from rsyscall.near.types import WatchDescriptor
from rsyscall.thread import Thread
import enum
import math
import rsyscall.handle as handle
import trio
import typing as t

from rsyscall.sys.inotify import InotifyFlag, IN, InotifyEvent, InotifyEventList

__all__ = [
    'Watch',
    'Inotify',
]

@dataclass
class Watch:
    "An indidivual inode being watched with an Inotify instance"
    inotify: Inotify
    wd: WatchDescriptor
    pending_events: t.List[InotifyEvent]

    def __init__(self, inotify: Inotify, wd: WatchDescriptor) -> None:
        self.inotify = inotify
        self.wd = wd
        self.pending_events = []
        self.queue = RequestQueue[None, t.List[InotifyEvent]]()
        reset(self._run())

    async def _run(self) -> None:
        waiters: t.List[t.Tuple[None, Continuation[t.List[InotifyEvent]]]] = []
        while True:
            received = await self.inotify.queue.request(self.wd)
            self.pending_events.extend(received)
            waiters.extend(self.queue.fetch_any())
            if waiters:
                _, to_resume = waiters.pop(0)
                to_send = self.pending_events
                self.pending_events = []
                to_resume.send(to_send)
            if any(event.mask & IN.IGNORED for event in received):
                # this watch was removed, we won't get any more events.
                break
        for _, waiter in waiters:
            waiter.throw(Exception("watch was removed"))
        waiters = []
        while True:
            _, waiter = await self.queue.get_one()
            waiter.throw(Exception("watch was removed"))

    async def wait(self) -> t.List[InotifyEvent]:
        "Wait for some events to happen at this inode"
        if self.pending_events:
            ret = self.pending_events
            self.pending_events = []
            return ret
        else:
            return await self.queue.request(None)

    async def wait_until_event(self, mask: IN, name: t.Optional[str]=None) -> InotifyEvent:
        """Wait until an event in this mask, and possibly with this name, happens

        Discards non-matching events.

        """
        while True:
            events = await self.wait()
            for event in events:
                if event.mask & mask and (event.name == name if name else True):
                    return event

    async def remove(self) -> None:
        "Remove this watch from inotify"
        await self.inotify.asyncfd.handle.inotify_rm_watch(self.wd)
        # we'll mark this Watch as removed once we get the IN_IGNORED event;
        # only after that do we know for sure that there are no more events
        # coming for this Watch.


_inotify_read_size = 4096
assert _inotify_read_size > InotifyEvent.MINIMUM_SIZE_TO_READ_ONE_EVENT

class Inotify:
    "An inotify file descriptor, which allows monitoring filesystem paths for events."
    def __init__(self, asyncfd: AsyncFileDescriptor, ram: RAM,
                 buf: Pointer[InotifyEventList],
    ) -> None:
        "Private; use Inotify.make instead."
        self.asyncfd = asyncfd
        self.ram = ram
        self.buf = buf
        self.wd_to_watch: t.Dict[WatchDescriptor, Watch] = {}
        self.queue = RequestQueue[WatchDescriptor, t.List[InotifyEvent]]()
        reset(self._run())

    @staticmethod
    async def make(thread: Thread) -> Inotify:
        "Create an Inotify file descriptor in `thread`."
        asyncfd = await AsyncFileDescriptor.make(
            thread.epoller, thread.ram, await thread.task.inotify_init(InotifyFlag.NONBLOCK))
        buf = await thread.ram.malloc(InotifyEventList, _inotify_read_size)
        return Inotify(asyncfd, thread.ram, buf)

    async def add(self, path: handle.Path, mask: IN) -> Watch:
        """Start watching a given path for events in the passed mask

        Note that if we monitor the same inode twice (whether at the same path or not),
        we'll return the same Watch object. Not sure how to make this usable.

        """
        wd = await self.asyncfd.handle.inotify_add_watch(await self.ram.ptr(path), mask)
        # if watch descriptors wrap, we could get back a watch descriptor that has been
        # freed and reallocated but for which we haven't yet read the IN.IGNORED event, so
        # we'd return the wrong Watch. but as the manpage says, that bug is very unlikely,
        # so the kernel has no mitigation for it; so we won't worry either.
        try:
            watch = self.wd_to_watch[wd]
        except KeyError:
            watch = Watch(self, wd)
            self.wd_to_watch[wd] = watch
        return watch

    async def _run(self) -> None:
        wd_to_cb: t.Dict[WatchDescriptor, Continuation] = {}
        while True:
            valid, rest = await self.asyncfd.read(self.buf)
            if valid.size() == 0:
                raise Exception('got EOF from inotify fd? what?')
            events = await valid.read()
            results: t.Dict[WatchDescriptor, t.List[InotifyEvent]] = {}
            for event in events:
                results.setdefault(event.wd, []).append(event)
            for wd, cb in self.queue.fetch_any():
                wd_to_cb[wd] = cb
            for wd, events in results.items():
                wd_to_cb[wd].send(events)
                del wd_to_cb[wd]
            self.buf = valid + rest



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread

from rsyscall import Pointer
from rsyscall.path import Path
from rsyscall.tests.utils import do_async_things
from rsyscall.fcntl import O, AT
from rsyscall.unistd import SEEK
from rsyscall.sched import CLONE
from rsyscall.stdlib import mkdtemp
from rsyscall.linux.dirent import *
from rsyscall.environ import ExecutablePathCache, ExecutableNotFound

class TestFS(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread
        self.tmpdir = await mkdtemp(self.thr)

    async def asyncTearDown(self) -> None:
        await self.tmpdir.cleanup()

    async def test_copy(self) -> None:
        source_file = await self.thr.task.open(await self.thr.ram.ptr(self.tmpdir/"source"), O.RDWR|O.CREAT)
        data = b'hello world'
        buf: Pointer[bytes] = await self.thr.ram.ptr(data)
        valid, rest = await source_file.write(buf)
        buf = valid + rest
        await source_file.lseek(0, SEEK.SET)
        dest_file = await self.thr.task.open(await self.thr.ram.ptr(self.tmpdir/"dest"), O.RDWR|O.CREAT)

        thread = await self.thr.clone()
        cat = await self.thr.environ.which("cat")
        await thread.task.inherit_fd(source_file).dup2(thread.stdin)
        await thread.task.inherit_fd(dest_file).dup2(thread.stdout)
        child_process = await thread.exec(cat)
        await child_process.check()

        await dest_file.lseek(0, SEEK.SET)
        self.assertEqual(await (await dest_file.read(buf))[0].read(), data)

    async def test_getdents(self) -> None:
        dirfd = await self.thr.task.open(await self.thr.ram.ptr(self.tmpdir), O.DIRECTORY)
        dent_buf = await self.thr.ram.malloc(DirentList, 4096)
        valid, rest = await dirfd.getdents(dent_buf)
        self.assertCountEqual([dirent.name for dirent in await valid.read()], ['.', '..'])
        dent_buf = valid + rest

        text = b"Hello world!"
        name = await self.thr.ram.ptr("hello")

        write_fd = await dirfd.openat(name, O.WRONLY|O.CREAT)
        buf = await self.thr.ram.ptr(text)
        written, _ = await write_fd.write(buf)

        read_fd = await dirfd.openat(name, O.RDONLY)
        read, _ = await read_fd.read(written)
        self.assertEqual(await read.read(), text)

        await dirfd.lseek(0, SEEK.SET)
        valid, rest = await dirfd.getdents(dent_buf)
        self.assertCountEqual([dirent.name for dirent in await valid.read()], ['.', '..', name.value])

    async def test_getdents_noent(self) -> None:
        "getdents on a removed directory returns ENOENT/FileNotFoundError"
        name = await self.thr.ram.ptr(self.tmpdir/"foo")
        await self.thr.task.mkdir(name)
        dirfd = await self.thr.task.open(name, O.DIRECTORY)
        await self.thr.task.rmdir(name)
        buf = await self.thr.ram.malloc(DirentList, 4096)
        with self.assertRaises(FileNotFoundError):
            await dirfd.getdents(buf)

    async def test_readlinkat_non_symlink(self) -> None:
        f = await self.thr.task.open(await self.thr.ptr("."), O.PATH)
        empty_ptr = await self.thr.ram.ptr("")
        ptr = await self.thr.ram.malloc(str, 4096)
        with self.assertRaises(FileNotFoundError):
            await f.readlinkat(empty_ptr, ptr)

    async def test_fdat(self) -> None:
        "The *at system calls on FileDescriptor work"
        root = await self.thr.task.open(await self.thr.ptr(self.tmpdir), O.DIRECTORY)
        name = await self.thr.ptr("foo")
        await root.mkdirat(name)
        await root.rmdirat(name)
        file = await root.openat(await self.thr.ptr("."), O.TMPFILE|O.WRONLY)
        await self.thr.task.linkat(None, await self.thr.ptr(file.as_proc_path()), root, name, AT.SYMLINK_FOLLOW)
        await root.renameat(name, root, name)
        await root.unlinkat(name)
        await root.symlinkat(name, name)

    async def test_readlink_proc(self) -> None:
        f = await self.thr.task.open(await self.thr.ptr("."), O.PATH)
        path_ptr = await self.thr.ram.ptr(f"/proc/self/fd/{int(f)}")
        ptr = await self.thr.ram.malloc(str, 4096)
        await f.readlinkat(path_ptr, ptr)

    async def test_which(self) -> None:
        names = []
        for i in range(5):
            name = await self.thr.ram.ptr(self.tmpdir/f"dir{i}")
            names.append(name)
            await self.thr.task.mkdir(name)
        cache = ExecutablePathCache(self.thr.task, self.thr.ram, [str(name.value) for name in names])

        with self.assertRaises(ExecutableNotFound):
            await cache.which("foo")
        with self.assertRaises(ExecutableNotFound):
            await cache.which("foo")
        fd = await self.thr.task.open(await self.thr.ram.ptr(names[2].value/"foo"), O.CREAT)
        await fd.fchmod(0o700)
        await cache.which("foo")
        



from __future__ import annotations
from dataclasses import dataclass
from dneio import Continuation, shift, reset
from dneio.outcome import Outcome
import functools
import logging
import outcome
import typing as t

logger = logging.getLogger(__name__)

InType = t.TypeVar('InType')
OutType = t.TypeVar('OutType')
T = t.TypeVar('T')

class RequestQueue(t.Generic[InType, OutType]):
    def __init__(self) -> None:
        self._request_cbs: t.List[t.Tuple[InType, Continuation[OutType]]] = []
        self._receiver_cb: t.Optional[Continuation[t.Tuple[InType, Continuation[OutType]]]] = None
        self._final_exn: t.Optional[BaseException] = None

    def close(self, final_exn: BaseException) -> None:
        self._final_exn = final_exn
        request_cbs, self._request_cbs = self._request_cbs, []
        for _, cb in request_cbs:
            cb.throw(self._final_exn)

    def request_cb(self, val: InType, cb: Continuation[OutType]) -> None:
        if self._final_exn:
            logger.debug("RequestQueue.request_cb(%s, %s): throwing final exn %s", val, cb, self._final_exn)
            cb.throw(self._final_exn)
        elif self._receiver_cb:
            logger.debug("RequestQueue.request_cb(%s, %s): waking up receiver_cb %s", val, cb, self._receiver_cb)
            receiver_cb = self._receiver_cb
            self._receiver_cb = None
            receiver_cb.send((val, cb))
        else:
            logger.debug("RequestQueue.request_cb(%s, %s): appending to waiting list of size %d", val, cb, len(self._request_cbs))
            self._request_cbs.append((val, cb))

    async def request(self, val: InType) -> OutType:
        if self._final_exn:
            raise self._final_exn
        logger.debug("RequestQueue.request(%s): shifting into request_cb", val)
        return await shift(functools.partial(self.request_cb, val))

    def get_one_cb(self, cb: Continuation[t.Tuple[InType, Continuation[OutType]]]) -> None:
        assert self._receiver_cb is None
        self._receiver_cb = cb

    async def get_one(self) -> t.Tuple[InType, Continuation[OutType]]:
        if self._request_cbs:
            return self._request_cbs.pop(0)
        else:
            return await shift(self.get_one_cb)

    async def get_many(self) -> t.List[t.Tuple[InType, Continuation[OutType]]]:
        if self._request_cbs:
            ret, self._request_cbs = self._request_cbs, []
            return ret
        else:
            return [await shift(self.get_one_cb)]

    def fetch_any(self) -> t.List[t.Tuple[InType, Continuation[OutType]]]:
        ret, self._request_cbs = self._request_cbs, []
        return ret

class Event:
    def __init__(self) -> None:
        self._waiting_cbs: t.List[Continuation[None]] = []
        self._is_set = False
        self._exc: t.Optional[BaseException] = None

    async def wait(self) -> None:
        if not self._is_set:
            await shift(self._waiting_cbs.append)
        if self._exc:
            raise self._exc

    def set(self) -> None:
        self._is_set = True
        for cb in self._waiting_cbs:
            cb.send(None)

    def close(self, exc: BaseException) -> None:
        self._exc = exc
        self.set()

@dataclass
class Future(t.Generic[T]):
    _result: t.Optional[Outcome[T]] = None
    _result_cb: t.Optional[Continuation[T]] = None

    @staticmethod
    def start(func: t.Callable[[], t.Awaitable[T]]) -> Future[T]:
        self = Future[T]()
        @functools.wraps(func)
        async def wrapper() -> None:
            result = await outcome.acapture(func)
            self._result = result
            if self._result_cb:
                self._result_cb.resume(result)
        reset(wrapper())
        return self

    def get_cb(self, cb: Continuation[T]) -> None:
        if self._result:
            return cb.resume(self._result)
        assert self._result_cb is None
        self._result_cb = cb

    async def get(self) -> T:
        if self._result:
            return self._result.unwrap()
        return (await shift(self.get_cb))

async def make_n_in_parallel(make: t.Callable[[], t.Awaitable[T]], count: int) -> t.List[T]:
    "Call `make` n times in parallel, and return all the results."
    return [await fut.get() for fut in
            [Future.start(make) for _ in range(count)]]

async def run_all(callables: t.List[t.Callable[[], t.Awaitable[T]]]) -> t.List[T]:
    "Call all the functions passed to it, and return all the results."
    return [await fut.get() for fut in
            [Future.start(func) for func in callables]]
